/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met: 
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <stdlib.h>
#include <stdio.h>

#include "stack.h"

call_stack_t *stack_push( call_stack_t *s, uint32_t addr )
{
    call_stack_t * new_s = (call_stack_t *) malloc( sizeof(call_stack_t) );
	
    new_s->next = s;
    new_s->addr = addr;
	
    return new_s;
}

call_stack_t *stack_pop( call_stack_t *s, uint32_t *addr )
{
    call_stack_t *ret;
	
    if ( s == NULL )
        return NULL;
	
    if ( addr != NULL )
        *addr = s->addr;
	
    ret = s->next;
    free( s );
	
    return ret;
}

void stack_dump( uint32_t pc, call_stack_t *s, int chipid )
{
    call_stack_t *c;
    //VMNet rem+++++++++++++++++++++++++
    return;                             
	//++++++++++++++++++++++++++++++++++
    printf( "Call stack %d begin:\n", chipid );
    printf( "\taddr: %08x\n", pc * 2 );
	
    for( c = s; c != NULL; c = c->next )
        printf( "\taddr: %08x\n", c->addr * 2 );
	
    fflush(stdout);
}

