/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include "../EM/atmega128.h"
#include "opcodes.h"
#include "../EM/debug.h"
#include "../EM/iomanager.h"
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//CPU operations

char status_cpu[MAX_NUM_LOGS]={-1};
unsigned char strCPUOperations[CPU_OPERATION_TYPES][MAX_OPERATION_STR_LEN]={"CPU:SLEEP","CPU:ACTIVE","CPU:FLASHW"};
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
error_t atmega_op_sleep( atmega_chip *chip )
{
    uint8_t mcucr = 0, sleepmode;
	
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip, "SLEEP\n");
        return 1;
    }
	
    chip->cycles += 1;
    chip->pc += 1;
    atmega_read_sram(chip, 0x0055, &mcucr);
    if(GET_FLAG(mcucr, 0x20)) {
        sleepmode = (mcucr >> 2) & 0x07;
        //sleepmode=0;//VMNet change here to disable hibernate.
        if((sleepmode == SLEEPMODE_RESERVED1)
			|| (sleepmode == SLEEPMODE_RESERVED2)) {
            log_error(chip, "Attempt to enter a reserved sleep mode failed.\n");
            return 0;
        }
		
        chip->am_sleeping = 1;
        atmega_ioman_enter_sleep(chip, sleepmode);
    } else {
        log_error(chip, "SLEEP instruction hit without Sleep Enable set.\n");
    }      
	// VMNet EM log +++++++++++++++++++++++++++++++++
    if(sleepmode)
	{
		//printf("%d: SLEEPT: %d SLEEP MODE:%d \n", chip->id, mcucr,sleepmode);
		LOG_CURRENT_OPERATION((uint16_t)(sleepmode<<12),0xf000)
		//g_processor_state_table[g_idChip].state = CPU_HIBERNATE;
		//exit(1);
	}
    else
		LOG_CURRENT_OPERATION(CPU_SLEEP,CPU_MASK);
	// VMNet EM log +++++++++++++++++++++++++++++++++    
	
    return 0;
}

error_t atmega_op_nop( atmega_chip *chip )
{
    if(chip->flash[chip->pc] != 0x0000){
        /* it's not really a true NOP */
        return -1;
    }
	
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "NOP\n" );
        return 1;
    }
	
    chip->cycles += 1;
    chip->pc += 1;
    
    return 0;
}

error_t atmega_op_wdr( atmega_chip *chip )
{
    log_error(chip,  "executing %d.\n", g_idChip );
    //run to unknown...
    return 0;
}

error_t atmega_op_break( atmega_chip *chip )
{
    if(chip->mode == ATEMU_MODE_DECODE){
        log_opcode(chip,  "BREAK\n" );
        return 1;
    }
	
	//    chip->cycles += 1;
	//    chip->pc += 1;
    
    return 0;
}
