/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMIlogsTED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: POWER_CONSUMPTION_MEASURE
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

/*
* Modifications:
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

  * Modification author: Hejun Wu (vmnet@cs.ust.hk)
  
********************************************************************************/

#include "usart.h"
#include "../EM/debug.h"
#include "../proxy/bridge.h"
#include <fcntl.h>
#include <unistd.h>

/* Local Static Functions */
static void usart_destroy_device(void *device);
static error_t usart_enter_sleep(void *device, uint8_t sleepmode);
static error_t usart_leave_sleep(void *device);
static int usart_clearint_rx(void *device);
static int usart_clearint_tx(void *device);
static int usart_clearint_udre(void *device);
static uint8_t usart_data_read(void *dev);
static void usart_data_write(void *dev, uint8_t value);
static uint8_t usart_controlA_read(void *dev);
static void usart_controlA_write(void *dev, uint8_t value);
static uint8_t usart_controlB_read(void *dev);
static void usart_controlB_write(void *dev, uint8_t value);
static uint8_t usart_controlC_read(void *dev);
static void usart_controlC_write(void *dev, uint8_t value);
static uint8_t usart_baudlow_read(void *dev);
static void usart_baudlow_write(void *dev, uint8_t value);
static uint8_t usart_baudhigh_read(void *dev);
static void usart_baudhigh_write(void *dev, uint8_t value);
static int usart_clock_transmit(void *dev);
static void usart_setint_rx(usart_device *usart);
static void usart_setint_tx(usart_device *usart);
static void usart_setint_udre(usart_device *usart);
static void usart_updatebaud(usart_device *usart);
/////////////////////////////////////////////////////////////////////
//VMNet UART global variable for connect to the proxy
extern unsigned char UDR[];
extern usart_device* g_usart;
uint8_t DataoutPacket[MTU_SIZE_UART];  
usart_device *g_moteUart[2];
static motecounter=0;


/////////////////////////////////////////////////////////////////////
/*
* Startup functions
*/
device_obj* create_device(char *name) {
	
    usart_device *new_dev = (usart_device*)malloc(sizeof(usart_device));
    if(new_dev == NULL) {
        return NULL;
    }
    g_moteUart[motecounter++] = new_dev;
	
    if(!g_usart) g_usart=new_dev;
    memset(new_dev, 0, sizeof(usart_device));
    /* Assume a 8MHz system clock by default */
    new_dev->fosc = 8000000;
    new_dev->filename = strdup("null");
    new_dev->file = -1;
    new_dev->filemode = O_WRONLY;
	
    new_dev->controlA = USART_CONTROLA_UDRE0;
    new_dev->controlB = 0;
    new_dev->controlC = USART_CONTROLC_UCSZ0_8BIT;
    new_dev->baud = 0;
    new_dev->baud = 16;
	
    new_dev->device.destroy = usart_destroy_device;
    new_dev->device.sleep = usart_enter_sleep;
    new_dev->device.wake = usart_leave_sleep;
    new_dev->device.dev = new_dev;
    new_dev->clk_tick.tick = NULL;
    new_dev->clk_tick.dev = new_dev;
    new_dev->int_rx.clear = usart_clearint_rx;
    new_dev->int_rx.dev = new_dev;
    new_dev->int_tx.clear = usart_clearint_tx;
    new_dev->int_tx.dev = new_dev;
    new_dev->int_udre.clear = usart_clearint_udre;
    new_dev->int_udre.dev = new_dev;
	
    new_dev->reg_data.read = usart_data_read;
    new_dev->reg_data.write = usart_data_write;
    new_dev->reg_data.dev = new_dev;
    new_dev->reg_controlA.read = usart_controlA_read;
    new_dev->reg_controlA.write = usart_controlA_write;
    new_dev->reg_controlA.dev = new_dev;
    new_dev->reg_controlB.read = usart_controlB_read;
    new_dev->reg_controlB.write = usart_controlB_write;
    new_dev->reg_controlB.dev = new_dev;
    new_dev->reg_controlC.read = usart_controlC_read;
    new_dev->reg_controlC.write = usart_controlC_write;
    new_dev->reg_controlC.dev = new_dev;
    new_dev->reg_baudhigh.read = usart_baudhigh_read;
    new_dev->reg_baudhigh.write = usart_baudhigh_write;
    new_dev->reg_baudhigh.dev = new_dev;
    new_dev->reg_baudlow.read = usart_baudlow_read;
    new_dev->reg_baudlow.write = usart_baudlow_write;
    new_dev->reg_baudlow.dev = new_dev;
	
    return &(new_dev->device);
}

register_obj *add_register(void *dev, char *name) {
    usart_device *usart = (usart_device*)dev;
	
    if(strcasecmp(name, "DATA") == 0) {
        return &(usart->reg_data);
    } else if(strcasecmp( name, "CONTROLA" ) == 0) {
        return &(usart->reg_controlA);
    } else if(strcasecmp( name, "CONTROLB" ) == 0) {
        return &(usart->reg_controlB);
    } else if(strcasecmp( name, "CONTROLC" ) == 0) {http://www.cshcn.umd.edu/research/atemu/downloads.html
	return &(usart->reg_controlC);
    } else if(strcasecmp( name, "BAUDLOW" ) == 0) {
        return &(usart->reg_baudlow);
    } else if(strcasecmp( name, "BAUDHIGH" ) == 0) {
        return &(usart->reg_baudhigh);
    }
	
    return NULL;
}

interrupt_obj* add_interrupt(void *dev, char *name) {
    usart_device *usart = (usart_device*)dev;
	
    if(strcasecmp(name, "RX") == 0) {
		
        return &(usart->int_rx);
    } else if(strcasecmp(name, "TX") == 0) {
        return &(usart->int_tx);
    } else if(strcasecmp(name, "UDRE") == 0) {
        return &(usart->int_udre);
    }
	
    return NULL;
}

clock_obj* add_clock(void *dev) {
    usart_device *timer = (usart_device*)dev;
    return &(timer->clk_tick);
}

error_t set_param(void *dev, char *key, char *value) {
    usart_device *usart = (usart_device*)dev;
    char *endparse;
    if(strcasecmp(key, "fosc") == 0) {
        uint32_t parse = strtoul(value, &endparse, 0);
        if(endparse && *endparse) {
            /* Non-numeric character */
            return -1;
        }
        usart->fosc = parse;
        return 0;
    } else if(strcasecmp(key, "filename") == 0) {
        if(usart->filename != NULL) {
			
            free(usart->filename);
            usart->filename = NULL;
        }
        /* If a file has been openned and not std*, close it */
        if(usart->file > 2) {
			
            close(usart->file);
        }
        usart->file = -1;
        /* If the user provided an empty string, make it NULL */
        if(strlen(value) == 0) {
            value = "null";
        }
        usart->filename = strdup(value);
        /* Handle special cases */
        if(strcasecmp(value, "null") == 0) {
            /* No operation */
        } else if(strcasecmp(value, "stdout") == 0) {
			
            usart->file = 1;
        } else if(strcasecmp(value, "stdin") == 0) {
			
            usart->file = 0;
        } else if(strcasecmp(value, "stderr") == 0) {
			
            usart->file = 2;
        } else {
            usart->file = open(value, usart->filemode, 0666);
            if(usart->file == -1) {
				
                fprintf(stderr, "Unable to open file %s: %s\n", value,
					strerror(errno));
                return -1;
            }
        }
		
        return 0;
    } else if(strcasecmp(key, "filemode") == 0) {
        /* File Mode */
        char *keyword = strtok(value, ", :&|+");
        int read = 1, write = 1, append = 0;
		
        /* Break down list of keywords */
        while(keyword != NULL) {
            if(strcasecmp(keyword, "read") == 0) {
				
                read = 1;
            } else if(strcasecmp(keyword, "write") == 0) {
				
                write = 1;
            } else if(strcasecmp(keyword, "append") == 0) {
				
                append = 1;
            } else {
                fprintf(stderr, "Unknown file mode keyword: %s\n", keyword);
            }
            keyword = strtok(NULL, ", :&|+");
        }
		
        /* Convert keywords to open() constants */
        if(read && write) {
            usart->filemode = O_RDWR;
			
        } else if(read) {
            usart->filemode = O_RDONLY;
			
        } else if(write) {
            usart->filemode = O_WRONLY;
			
        } else {
            fprintf(stderr, "File mode does not include reading or writing!\n");
			
            usart->filemode = O_RDWR;
        }
        if(append) {
			
            usart->filemode |= O_APPEND;
        } else if(!read) {
			
            usart->filemode |= O_TRUNC;
        }
        if(write) {
			
            usart->filemode |= O_CREAT;
        }
		
        /* Store summary for transmit and receive functions */
        usart->canread = read;
        usart->canwrite = write;
		
        return 0;
    }
	
	
    return -1;
}

/*
* Basic device callbacks
*/

static void usart_destroy_device(void *dev) {
    usart_device *usart = (usart_device*)dev;
	
    if(usart->filename) {
        free(usart->filename);
        usart->filename = NULL;
    }
    if(usart->file > 2) {
        close(usart->file);
        usart->file = -1;
    }
    free(usart);
}

static error_t usart_enter_sleep(void *dev, uint8_t sleepmode) {
    usart_device *usart = (usart_device*)dev;
	
    switch(sleepmode) {
    case SLEEPMODE_IDLE:
        /* Continue to operate in idle */
        break;
    case SLEEPMODE_ADCNOISE:
    case SLEEPMODE_EXTSTANDBY:
    case SLEEPMODE_POWERDOWN:
    case SLEEPMODE_POWERSAVE:
    case SLEEPMODE_STANDBY:
        /* Disabled in all other sleep modes */
        device_clrclock(usart->device, usart->clk_tick);
        if(usart->datadelay) {
            usart->device.device_debug(&usart->device, V_NORMAL,
				"Going to sleep while transmitting data!\n");
        }
        break;
    default:
        /* Shouldn't be in this mode.... */
        usart->device.device_debug(&usart->device, V_NORMAL,
			"Unknown sleep mode %d!\n", sleepmode);
        device_clrclock(usart->device, usart->clk_tick);
        return -1;
    }
	
    return 0;
}

static error_t usart_leave_sleep(void *dev) {
    usart_device *usart = (usart_device*)dev;
	
    /* Restart clock if outgoing data */
    if(!GET_FLAG(usart->controlA, USART_CONTROLA_UDRE0)
		|| (usart->datadelay)) {
        device_setclock(usart->device, usart->clk_tick, usart_clock_transmit);
    }
    return 0;
}

static int usart_clearint_rx(void *dev) {
    usart_device *usart = (usart_device*)dev;
    usart->controlA &= ~USART_CONTROLA_RXC0;
    return 0;
}

static int usart_clearint_tx(void *dev) {
    usart_device *usart = (usart_device*)dev;
    usart->controlA &= ~USART_CONTROLA_TXC0;
	
    return 0;
}

static int usart_clearint_udre(void *dev) {
    usart_device *usart = (usart_device*)dev;
	
    /* This interrupt will not go away until masked or dealt with */
    if(GET_FLAG(usart->controlA, USART_CONTROLA_UDRE0)) {
		
        usart_setint_udre(usart);
    }
	
    return 0;
}

/*
* Device I/O
*/
///////////////////////////////////////////////////////////////////
//VMNet UART read
//////////////////////////////////////
#if 0
#define QUERY_POST_SIZE 52
#define QUERY_POSTED_TIME 2
static uint8_t usart_data_read(void *dev) {
    /* VMNet need this kind implementation */
    usart_device *usart = (usart_device*)dev;
    static int pointer =0;
    static int received=0;
    uint8_t b=UDR[pointer];
	
	
	//    DEBUG_OUTPUT("CURRENT DATA IS %x \n",b);
    if(pointer==0&&b!=SYNC_BYTE)
    {
		
		return 0;
    }
    if(b!=SYNC_BYTE)
		usart_setint_rx(usart);
    else
    {
#ifdef LOG_INITIALIZATION_TIME      
        if(pointer==QUERY_POST_SIZE)
          received++;
        printf("received =====%d  %d msg\n",pointer,received);
        if(received>=QUERY_POSTED_TIME)
        {

            int i=0;
            g_log_controllreg=LOG_CONTROL_START;
        
            while(i<g_nChips)
				        	{
          					char tempFileName[80]={0};
                    if(g_power_log[i])
                     fclose(g_power_log[i]);
                      printf("log new\n");
			          			sprintf(tempFileName,LOG_FILE_STR,g_query_id,i,0);
          						g_power_log[i]=fopen(tempFileName, "w");
					          	printf("log file name %s\n",tempFileName);
          						if (g_power_log[i] == NULL) {
          						  perror("Error opening write pipe file");
					          	  exit(-1);
      				    	  }
                    i++;
					        }

        }
       #endif 
        if(pointer)
       { pointer=0; UDR[0]=0;}
	else
        usart_setint_rx(usart);
    }
    pointer++;
    return b;
    
    return 0;
}
#endif
///////////////////////////////////////////////////////////////////
//VMNet UART read
//////////////////////////////////////
#define QUERY_POST_SIZE 52
#define QUERY_POSTED_TIME 2
#define USART_WAIT_COUNT  30
static uint8_t usart_data_read(void *dev) 
{
    /* VMNet need this kind implementation */
    usart_device *usart = (usart_device*)dev;
    static int pointer =0;
    static int received=0;
    static int n_wait_count=0, n_wait_timer=USART_WAIT_COUNT;
    uint8_t b=UDR[pointer];
    if(pointer==0&&b!=SYNC_BYTE)
		return 0;
    if(b!=SYNC_BYTE)
		usart_setint_rx(usart);
    else
    {
#ifdef LOG_INITIALIZATION_TIME      
        if(pointer==QUERY_POST_SIZE)
			received++;
        //printf("received =====%d  %d msg\n",pointer,received);
        if(received>=QUERY_POSTED_TIME)
        {
            int i=0;
            g_log_controllreg=0;
	    perror("Restart log...\n");
            while(i<g_nChips)
			{
          		char tempFileName[80]={0};
                if(g_power_log[i])
				{
					fclose(g_power_log[i]);
					g_power_log[i]=0;
					}
               
				sprintf(tempFileName,LOG_FILE_STR,g_query_id,i,0);
				g_power_log[i]=fopen(tempFileName, "w");
				fprintf(stderr,"log file name %s\n",tempFileName);
				if (g_power_log[i] == NULL) 
				{
					perror("Error opening write pipe file");
					exit(-1);
				}
				i++;
			}
		}
#endif 
		if(pointer)
			{ pointer=0; UDR[0]=0;}
		else
		{
		
			if(pointer==0)
				n_wait_timer+=4;
			if((n_wait_count++) % n_wait_timer ==0)
			{
				n_wait_timer = USART_WAIT_COUNT;
				usart_setint_rx(usart);
			}
		}
    }
    pointer++;
    return b;
}
///////////////////////////////////////////////////////////////////
static void usart_data_write(void *dev, uint8_t value) {
    usart_device *usart = (usart_device*)dev;
    uint16_t wordout = value;
	
    /* Data writes are ignored when UDREn is not set (buffer is full) */
    if(!GET_FLAG(usart->controlA, USART_CONTROLA_UDRE0)) {
		
        return;
    }
    /* Append bit 9 if relavent */
    if(GET_FLAG(usart->controlB, USART_CONTROLB_TXB80)) {
        wordout |= 0x0100;
    }
    /* Store value and set up clock to transmit when appropriate */
	
    usart->dataout = wordout;
	
    if(GET_FLAG(usart->controlB, USART_CONTROLB_TXEN0)) {
        device_setclock(usart->device, usart->clk_tick, usart_clock_transmit);
    }
    /* The data register is no longer empty */
    usart->controlA &= ~USART_CONTROLA_UDRE0;
    usart->int_udre.clear_interrupt(&usart->int_udre);
}
///////////////////////////////////////////////////////////////////
//VMNet UART read control A
//////////////////////////////////////

static uint8_t usart_controlA_read(void *dev) {
    usart_device *usart = (usart_device*)dev;
	
    if(UDR[0]==SYNC_BYTE)
    {
        usart->controlA|=USART_CONTROLA_RXC0;
	}
    return (usart->controlA);
}                          
/////////////////////////////////////////////////////////////////////

static uint8_t usart_controlB_read(void *dev) {
	
    usart_device *usart = (usart_device*)dev;
    return usart->controlB;
}

static uint8_t usart_controlC_read(void *dev) {
    usart_device *usart = (usart_device*)dev;
	
    return usart->controlC;
}

static void usart_controlA_write(void *dev, uint8_t value) {
    usart_device *usart = (usart_device*)dev;
    uint8_t newvalue = (value & ~USART_CONTROLA_STATUS)
        | (usart->controlA & USART_CONTROLA_STATUS);
    if(value & USART_CONTROLA_TXC0) {
        newvalue &= ~USART_CONTROLA_TXC0;
        usart->int_tx.clear_interrupt(&(usart->int_tx));
    }
    usart_updatebaud(usart);
}                                                   
///////////////////////////////////////////////////////////////////
//VMNet UART write B when ever the newvalue ready
//////////////////////////////////////
static void usart_controlB_write(void *dev, uint8_t value) {
    usart_device *usart = (usart_device*)dev;
    uint8_t newvalue = (value & ~USART_CONTROLB_STATUS)
        | (usart->controlB & USART_CONTROLB_STATUS);
    /* Set/clear interrupts if enabled/disabled */
    if(GET_FLAG(newvalue, USART_CONTROLB_RXCIE0)) {
        if(GET_FLAG(usart->controlA, USART_CONTROLA_RXC0)) {
			
			usart->controlB = newvalue;    //VMNet
            usart_setint_rx(usart);
        }
    } else {
		
        usart->int_rx.clear_interrupt(&usart->int_rx);
    }
    if(GET_FLAG(newvalue, USART_CONTROLB_TXCIE0)) {
        if(GET_FLAG(usart->controlA, USART_CONTROLA_TXC0)) {
            usart_setint_tx(usart);
        }
    } else {
        usart->int_tx.clear_interrupt(&usart->int_tx);
    }
    if(GET_FLAG(newvalue, USART_CONTROLB_UDRIE0)) {
        if(GET_FLAG(usart->controlA, USART_CONTROLA_UDRE0)) {
            usart_setint_udre(usart);
        }
    } else {
        usart->int_udre.clear_interrupt(&usart->int_udre);
    }
    /* If transmissions are being enabled, output them */
    if(GET_FLAG(newvalue, USART_CONTROLB_TXEN0)
		!= GET_FLAG(usart->controlB, USART_CONTROLB_TXEN0)) {
        if(GET_FLAG(newvalue, USART_CONTROLB_TXEN0)) {
            usart->device.device_debug(&usart->device, V_DEBUG,
				"Enabling transmission.\n");
            device_setclock(usart->device, usart->clk_tick,
				usart_clock_transmit);
        } else {
            if(usart->datadelay) {
                usart->device.device_debug(&usart->device, V_NORMAL,
					"Transmission disabled while sending data!\n");
            } else {
                usart->device.device_debug(&usart->device, V_DEBUG,
					"Disabling transmission.\n");
            }
            device_clrclock(usart->device, usart->clk_tick);
        }
    }
    usart->controlB = newvalue;
}                                                
///////////////////////////////////////////////////////

static void usart_controlC_write(void *dev, uint8_t value) {
    usart_device *usart = (usart_device*)dev;
    uint8_t newvalue = (value & ~USART_CONTROLC_STATUS)
        | (usart->controlC & USART_CONTROLC_STATUS);
    /* Notify about Sychronous/Asynchronous mode changes */
    if(GET_FLAG(newvalue, USART_CONTROLC_UMSEL0)
		!= GET_FLAG(usart->controlC, USART_CONTROLC_UMSEL0)) {
        if(GET_FLAG(newvalue, USART_CONTROLC_UMSEL0)) {
            usart->device.device_debug(&usart->device, V_NORMAL,
				"USART set to Synchronous Operation (not supported).\n");
        } else {
            usart->device.device_debug(&usart->device, V_DEBUG,
				"USART set to Asynchronous Operation.\n");
        }
    }
    /* Notify changes in parity */
    if((newvalue & USART_CONTROLC_UMP0)
		!= (usart->controlC & USART_CONTROLC_UMP0)) {
        switch(newvalue & USART_CONTROLC_UMP0) {
		case USART_CONTROLC_UMP0_DISABLE:
			usart->device.device_debug(&usart->device, V_DEBUG,
				"Parity disabled.\n");
			break;
		case USART_CONTROLC_UMP0_RESERVE:
			usart->device.device_debug(&usart->device, V_NORMAL,
				"Parity set to illegal reserved value!\n");
			break;
		case USART_CONTROLC_UMP0_EVEN:
			usart->device.device_debug(&usart->device, V_DEBUG,
				"Parity enabled and set to even parity.\n");
			break;
		case USART_CONTROLC_UMP0_ODD:
			usart->device.device_debug(&usart->device, V_DEBUG,
				"Parity enabled and set to odd parity.\n");
			break;
        }
    }
    /* Notify changes in stop bit count */
    if(GET_FLAG(newvalue, USART_CONTROLC_UMSEL0)
		!= GET_FLAG(usart->controlC, USART_CONTROLC_UMSEL0)) {
        usart->device.device_debug(&usart->device, V_DEBUG,
			"Using %d stop bit%s.\n",
			(GET_FLAG(newvalue, USART_CONTROLC_UMSEL0) ? 2 : 1),
			(GET_FLAG(newvalue, USART_CONTROLC_UMSEL0) ? "s" : ""));
    }
    /* Notify changes in character sizes */
    if((newvalue & USART_CONTROLC_UCSZ0)
		!= (usart->controlC & USART_CONTROLC_UCSZ0)) {
        if(!GET_FLAG(usart->controlB, USART_CONTROLB_UCXZ02)) {
            switch(newvalue & USART_CONTROLC_UCSZ0) {
			case USART_CONTROLC_UCSZ0_5BIT:
				usart->device.device_debug(&usart->device, V_DEBUG,
					"Character size set to 5 bits.\n");
				break;
			case USART_CONTROLC_UCSZ0_6BIT:
				usart->device.device_debug(&usart->device, V_DEBUG,
					"Character size set to 6 bits.\n");
				break;
			case USART_CONTROLC_UCSZ0_7BIT:
				usart->device.device_debug(&usart->device, V_DEBUG,
					"Character size set to 7 bits.\n");
				break;
			case USART_CONTROLC_UCSZ0_8BIT:
				usart->device.device_debug(&usart->device, V_DEBUG,
					"Character size set to 8 bits.\n");
				break;
            }
        } else if((newvalue & USART_CONTROLC_UCSZ0)
			!= USART_CONTROLC_UCSZ0_8BIT) {
            usart->device.device_debug(&usart->device, V_NORMAL,
				"Character size set to a reserved value!\n");
        } else if((usart->controlC & USART_CONTROLC_UCSZ0)
			!= USART_CONTROLC_UCSZ0_8BIT) {
            usart->device.device_debug(&usart->device, V_DEBUG,
				"Character size set to 9 bits.\n");
        }
    }
    /* Actually overright the old Control register C value */
    usart->controlC = newvalue;
	
}

static uint8_t usart_baudlow_read(void *dev) {
    usart_device *usart = (usart_device*)dev;
    return (uint8_t)(usart->baud & 0x00FF);
}

static uint8_t usart_baudhigh_read(void *dev) {
    usart_device *usart = (usart_device*)dev;
	
    return (uint8_t)(usart->baud >> 8);
}

static void usart_baudlow_write(void *dev, uint8_t value) {
    usart_device *usart = (usart_device*)dev;
    usart->baud = (usart->baud & 0xFF00) | (uint16_t)value;
    usart_updatebaud(usart);
}

static void usart_baudhigh_write(void *dev, uint8_t value) {
    usart_device *usart = (usart_device*)dev;
    usart->baud = (usart->baud & 0x00FF) | (((uint16_t)value) << 8);
    usart_updatebaud(usart);
}

///////////////////////////////////////////////////////////////////
//VMNet UART send data
//////////////////////////////////////

extern unsigned char g_tempCompareDbg;
extern char g_tempSubDbg;
extern unsigned short g_nChips;
static int s_ExpCounter = 0;
static int g_start_logs = 0;
static uint64_t g_endCycles;
static int nTotalChips[]={8,8,6,6,6,4,3};//whj temp
//static int nTotalChips[]={8,16,16,16,8,};//whj temp
#define ESCAPE_NO 0x7D
#define ESCAPE_PATCH 0x5D
int escape_finder(char* packet, int*offset1, int length)
{
	int i=1,flag=0;
	int offset=0;
	for(;i<length;i++)
	{
		if(packet[i]==ESCAPE_NO)
			offset++;
		if(i==(*offset1-1))
		{
			*offset1=offset;
			flag=1;
		}
	}
	if(!flag) 
	{fprintf(stderr,"ERROR PARAMETER!!!"); exit(-1);}
	return offset;
}
//#define TIME_LOG_ONLY
#ifdef POWER_CONSUMPTION_MEASURE
void single_node_logger (int pointer)
{
	static uint16_t s_count_epochs=0;
	int epoch_offset;
	int offset ;
	int current_epoch, current_mote;
	if(pointer < (MOTE_ID_LOCATION+1))
		return;
	epoch_offset = EPOCH_LOCATION_IN_PACKET;
	offset = escape_finder(DataoutPacket,&epoch_offset,MOTE_ID_LOCATION);
	current_epoch = DataoutPacket[EPOCH_LOCATION_IN_PACKET+epoch_offset];
	//printf("offset=======%d epochoffset==========%d\n",offset,epoch_offset);
	if(current_epoch==1)
		return;
	if(current_epoch==ESCAPE_NO)
	{
		if(DataoutPacket[EPOCH_LOCATION_IN_PACKET + epoch_offset+1 ]>=ESCAPE_PATCH)
			current_epoch+=(DataoutPacket[EPOCH_LOCATION_IN_PACKET+epoch_offset+1]-ESCAPE_PATCH);
	}
	// FOR ANY DATA DATA   
	
	current_mote = DataoutPacket[MOTE_ID_LOCATION+offset];      
	if(current_mote<g_nChips)
	{
		static char flagLog=0;
        static minEpochs=0;
		if(g_epoch_id[current_mote]<current_epoch)
		{
			g_epoch_id[current_mote]=current_epoch;
			#ifndef TIME_LOG_ONLY
			flagLog=1;
			#endif
			
		}
		else {if(current_epoch)fprintf(stderr,"BACK FRONT!!!! %d --- %d \n",g_epoch_id[current_mote], current_epoch); }
		if(!g_start_logs)
		{
			
			if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>LOG_CONTROL_START)
			{
				int nLogs=0;
		        char temp_file_name[100];
				g_start_logs=1;
				g_log_controllreg=LOG_CONTROL_START;
				g_startCycles=g_chips[0]->cycles;
				g_log_startreg_flag=1;
				flagLog=1;
				
				#ifndef TIME_LOG_ONLY
				#if 0
				while(nLogs<g_nChips)
				{
					if(g_power_log[nLogs])
						fclose(g_power_log[nLogs++]);
					g_power_log[nLogs-1]=0;
				}
				#endif 
				s_ExpCounter++;
				if(current_mote)
					g_total_results++;
			    sprintf(temp_file_name,FILE_EPOCHS_RESULTS_LOG,g_query_id);
				g_epoch_result_log=fopen(temp_file_name,"w");
				if(!g_epoch_result_log)
				{
					perror("ERROR in opening file for reading...\n");
					exit(-1);
				}
				fprintf(g_epoch_result_log, "start epochs: \n");
				for (nLogs=0;nLogs<g_nChips;nLogs++)
					fprintf(g_epoch_result_log, "%d \t",g_epoch_id[nLogs]);
				fprintf(g_epoch_result_log, "\n\n");
				nLogs=0;
				while(nLogs<g_nChips)
				{
					char tempFileName[80]={0};
					{
						sprintf(tempFileName,LOG_FILE_STR,g_query_id,nLogs,1);
						fprintf(stderr,"log file name %s\n",tempFileName);
						g_power_log[nLogs]=fopen(tempFileName, "w");						
						if (g_power_log[nLogs] == NULL)
						{
							perror("Error opening write pipe file xxxx");
							exit(-1);
						}
						
					}
					nLogs++;
				}
				if(g_epoch_id[current_mote])
					fprintf(g_power_log[current_mote], "EPOCH%d\n",g_epoch_id[current_mote]); 

				#else
				#endif
			}
		}
		else
		{
			if(current_mote)
				g_total_results++;
			if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>=(uint32_t)LOG_TOTAL_CYCLES)
			{
				int i=0;
				while(i<g_nChips)
				{
					if(g_power_log[i])
						fclose(g_power_log[i]);
					g_power_log[i]=0;
					i++;
				}
				g_log_controllreg=LOG_CONTROL_END;
				fprintf(g_epoch_result_log, "end epochs: \n");
				for (i=0;i<g_nChips;i++)
					fprintf(g_epoch_result_log, "%d \t",g_epoch_id[i]);
				fprintf(g_epoch_result_log, "\n\n");
				fprintf(g_epoch_result_log, "%d\n",g_total_results);
				if(g_epoch_result_log)
					fclose(g_epoch_result_log);				
				fprintf(stderr,"finished logging,  nomal exiting ... \n \n ");
				exit(0);
			}
			else{
				if(flagLog)
				{
					char tempFileName[80]={0};
					int i = 0;
					if(g_epoch_id[current_mote])
						fprintf(g_power_log[current_mote], "EPOCH%d\n",g_epoch_id[current_mote]);
				}
			}//else (g_chips[0]->cycles-g_startCycles)<=     
		}//else     
	}//if(
	else{ 
	if(g_start_logs)
		{perror("DATA ERROR!!!!!!!!!!!!!!!!!!\n");}
	}
}

void sink_logger(int pointer)
{
	int epoch_offset;
	int offset ;
	int current_epoch, current_mote;
	int nLogs=0;
	if(pointer < (EPOCH_LOCATION_IN_PACKET))
		return;
	epoch_offset = EPOCH_LOCATION_IN_PACKET;
	offset = escape_finder(DataoutPacket,&epoch_offset,MOTE_ID_LOCATION);
	current_epoch = DataoutPacket[EPOCH_LOCATION_IN_PACKET+epoch_offset];
	//printf("offset=======%d epochoffset==========%d\n",offset,epoch_offset);
	if(current_epoch==ESCAPE_NO)
	{
		if(DataoutPacket[EPOCH_LOCATION_IN_PACKET + epoch_offset+1 ]>=ESCAPE_PATCH)
			current_epoch+=(DataoutPacket[EPOCH_LOCATION_IN_PACKET+epoch_offset+1]-ESCAPE_PATCH);
	}
#ifdef LOG_AGGREGATION_INITIALIZATION
	if(pointer > (MOTE_ID_LOCATION+1)&&(DataoutPacket[QUERY_RESULT_LOCATION]==QUERY_RESULT_FLAG))
	{
		static uint16_t s_count_epochs=0;
		if(current_epoch>1)
			s_count_epochs++;
		if(!g_epoch_result_log)
		{
     		char temp_file_name[100];
     		sprintf(temp_file_name,FILE_EPOCHS_RESULTS_LOG,g_query_id);
			g_epoch_result_log=fopen(temp_file_name,"w");
			if(!g_epoch_result_log)
			{
			   	perror("ERROR in opening file for reading...\n");
				exit(-1);
			}
		}
		
		
			//if(g_log_controllreg<LOG_CONTROL_START)
			if(!g_start_logs)
			{
				if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>LOG_CONTROL_START)
			    {
					g_start_logs =1;
					g_startCycles=g_chips[0]->cycles;
					nLogs=0;
					g_log_controllreg=LOG_CONTROL_START;
					g_log_startreg_flag=1;
					while(nLogs<g_nChips)
					{
						char tempFileName[80]={0};
						
						sprintf(tempFileName,LOG_FILE_STR,g_query_id,nLogs,1);                        
						fprintf(stderr,"%d: log file name %s\n",current_epoch,tempFileName);
						g_power_log[nLogs]=fopen(tempFileName, "w");						
						if (g_power_log[nLogs] == NULL) {
							perror("Error opening write pipe file");
							exit(-1);
						}						
						fprintf(g_power_log[nLogs], "EPOCH%d\n",current_epoch);//log the first
						nLogs++;
						
					}
						fprintf(stderr,"log new\n");
                }
			}
				else {
						nLogs=0;

						while(nLogs<g_nChips)
						{
						fprintf(g_power_log[nLogs], "EPOCH%d\n",current_epoch);

						nLogs++;
						}
						nLogs=0;
						while(nLogs<pointer)
						{
							fprintf(g_epoch_result_log,"0x%02x,",DataoutPacket[nLogs]);
							nLogs++;
						}
						fprintf(g_epoch_result_log,"\n");

						if((g_chips[0]->cycles/7372800-g_startCycles/7372800)>=(uint32_t)LOG_TOTAL_CYCLES)
						{
							int i=0;
							while(i<g_nChips)
							{
								if(g_power_log[i])
									fclose(g_power_log[i]);
									i++;						
							}
							g_log_controllreg=LOG_CONTROL_END;
							if(g_epoch_result_log)
								fclose(g_epoch_result_log);
							fprintf(stderr,"finished logging, auto exit.... Remember to copy the logs!!\n");
							exit(0);
						}

            	    }
	    
		
	}
#else
	if(pointer > (MOTE_ID_LOCATION+1))
	{
		//for epochs
		int controlreg=0;
		if(REFERENCE_MOTE_ID)
		{
			if(DataoutPacket[MOTE_ID_LOCATION]==REFERENCE_MOTE_ID)
				controlreg=1;
		}
		else if(DataoutPacket[QUERY_RESULT_LOCATION]==QUERY_RESULT_FLAG)
				controlreg=1;
		if (!controlreg)
			return;
		if(g_log_controllreg<LOG_CONTROL_END)
			g_log_controllreg+=1;
		if(g_log_controllreg==LOG_CONTROL_END)
		{
			int nLogs=0;
			char tempLabel[10];
			g_log_controllreg=LOG_CONTROL_START-1;
			s_ExpCounter++;
			while(nLogs<g_nChips)
			{
			if(g_power_log[nLogs])
				fclose(g_power_log[nLogs++]);
				g_power_log[nLogs-1]=0;
			}
			if(s_ExpCounter<EXP_TIMES)
			{
				nLogs=0;
				while(nLogs<g_nChips)
				{
					char tempFileName[80]={0};
                    sprintf(tempFileName,LOG_FILE_STR,g_query_id,nLogs,s_ExpCounter);                        
					g_power_log[nLogs]=fopen(tempFileName, "w");
					fprintf(stderr,"log file name %s\n",tempFileName);
					if (g_power_log[nLogs] == NULL) {
						perror("Error opening write pipe file");
						exit(-1);
					}
					nLogs++;
				} 
			}
			else
			{
				exit(0);
			}
		}
	}

#endif 

}

#endif
static int usart_clock_transmit(void *dev) {
    usart_device *usart = (usart_device*)dev;
    static int pointer=0;
    /* Make sure we are still running */
    if(usart->datadelay > 0) {
        /* Currently "sending" a character */
        usart->datadelay--;
        if(usart->datadelay == 0) {
            /* Byte sent, check for interrupts */
            if(GET_FLAG(usart->controlA, USART_CONTROLA_UDRE0)) {
                /* Transmission complete */
                usart_setint_tx(usart);
            }
        }
        return 0;
    } else if(!GET_FLAG(usart->controlA, USART_CONTROLA_UDRE0)) {
        /* Finished previous character and a new one is waiting */
        if(usart->canwrite) {
            uint8_t dataout[2];
            int writeret;
			
            /* Construct actual outgoing byte */
            if(GET_FLAG(usart->controlB, USART_CONTROLB_UCXZ02)) {
                /* 9 Bits of data */
                dataout[0] = LOW_BYTE_16(usart->dataout);
                dataout[1] = (usart->dataout & 0x0100) ? 1 : 0;
				
                writeret = write(usart->file, dataout, 2);
            } else {
                /* 5, 6, 7, or 8 bits of data */
                switch(usart->controlC & USART_CONTROLC_UCSZ0) {
                case USART_CONTROLC_UCSZ0_8BIT:
                    dataout[0] = (usart->dataout & 0x00FF);
                    break;
                case USART_CONTROLC_UCSZ0_7BIT:
                    dataout[0] = (usart->dataout & 0x007F);
                    break;
                case USART_CONTROLC_UCSZ0_6BIT:
                    dataout[0] = (usart->dataout & 0x003F);
                    break;
                case USART_CONTROLC_UCSZ0_5BIT:
                    dataout[0] = (usart->dataout & 0x001F);
                    break;
                }
                {
					usart->controlA |= USART_CONTROLA_UDRE0;
					DataoutPacket[pointer++] = dataout[0];
					if(dataout[0]==SYNC_BYTE)
					{
						if ( pointer>1)
						{
							transmitEvent(0, 0, DataoutPacket, pointer);
							#ifdef POWER_CONSUMPTION_MEASURE
                            if(g_query_id<=SINGLE_NODE_LOG)
			    {
    							single_node_logger (pointer);
							
							}
                            else
                           		sink_logger(pointer);
					#endif
							pointer = 0;                       
						}
					}
                }
            } /* if(GET_FLAG(usart->controlB, USART_CONTROLB_UCSZ02)) { */
            /* Check for write errors */
            if(writeret == -1) {
                fprintf(stderr, "Error writing to %s: %s\n",
					usart->filename, strerror(errno));
                usart->canwrite = 0;
                if(usart->canread == 0) {
                    close(usart->file);
                    usart->file = -1;
                }
            }
        } /* if(usart->canwrite) { */
        /* Notify program the data register is empty and begin transmission */
        usart_setint_udre(usart);
        usart->datadelay = usart->databaud;
        return 0;
    } else {
        /* Buffer empty */
        device_clrclock(usart->device, usart->clk_tick);
        return 0;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////

static void usart_setint_rx(usart_device *usart) {
    usart->controlA |= USART_CONTROLA_RXC0;
    if(GET_FLAG(usart->controlB, USART_CONTROLB_RXCIE0)) {
        usart->int_rx.set_interrupt(&(usart->int_rx));
    }
}

static void usart_setint_tx(usart_device *usart) {
    usart->controlA |= USART_CONTROLA_TXC0;
    if(GET_FLAG(usart->controlB, USART_CONTROLB_TXCIE0)) {
        usart->int_tx.set_interrupt(&(usart->int_tx));
    }
}

static void usart_setint_udre(usart_device *usart) {
    usart->controlA |= USART_CONTROLA_UDRE0;
    if(GET_FLAG(usart->controlB, USART_CONTROLB_UDRIE0)) {
        usart->int_udre.set_interrupt(&(usart->int_udre));
    }
}

static void usart_updatebaud(usart_device *usart) {
    int16_t newbaud = usart->baud + 1;
    if(GET_FLAG(usart->controlA, USART_CONTROLA_U2X0)) {
        newbaud *= 8;
    } else {
        newbaud *= 16;
    }
    /* Print debugging messages */
    if(usart->databaud != newbaud) {
        if(usart->datadelay) {
            usart->device.device_debug(&usart->device, V_NORMAL,
				"Baud rate change while transmitting data!\n");
        }
        usart->device.device_debug(&usart->device, V_DEBUG,
			"Baud rate set to %d bits/second.\n", usart->fosc / newbaud);
    }
    usart->databaud = newbaud;
}
