 /*
* Modifications: 
*
* Copyright (c) 2004-2005, The Hong Kong University of Science and Technology (HKUST)
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
*   * Redistributions of source code must retain the above copyright notice,
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the university nor the names of its
*     contributors may be used to endorse or promote products derived from
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*
*/


/*

 * Modification author: Bingsheng He (saven@cs.ust.hk)
 
********************************************************************************/
/*******************************************************
 * This data analyzer is for VMNET logs analysis (non-sketch) only.
 * Author: Bingsheng He (saven@cs.ust.hk)
  * 2004-Nov-14.
 *******************************************************
 */

import java.sql.*;
import java.util.*;
import java.io.*;
class LogFilter implements FilenameFilter {
    public boolean accept(File dir, String name) {
        return (name.endsWith("epoch_result.txt"));
    }
}
class LogQueryFilter implements FilenameFilter {
    public boolean accept(File dir, String name) {
        return (name.indexOf("9")!=-1 || name.indexOf("6")!=-1 || name.indexOf("7")!=-1 ||name.indexOf("8")!=-1);
    }
}

class DatFilter implements FilenameFilter {
    public boolean accept(File dir, String name) {
        return ((name.indexOf("Sensor")!=-1) && name.endsWith(".dat"));
    }
}


public class QualityAnalyzer{
	private Connection con;
	public String aodvDir;
	public String tinyDBDir;
	public String DDDir;
	public int SP;
	public String rdDir;
	public String locFile;
	public Hashtable locHT;
	public int numGroups=3;
	public Hashtable idHT;

	QualityAnalyzer()
	{
		locHT=new Hashtable();
		idHT=new Hashtable();
	}
	public void buildLocHT()
	{
		try{
			BufferedReader in = new BufferedReader(new FileReader(locFile));
			String line=null;
			int x=0;
			int y=0;
			int groupid=1;
			int nodeid=0;
			int curGid=0;
			Hashtable tempHT=new Hashtable();
			while(true)
			{
				line=in.readLine();
				if(line==null) break;

				String []inner=line.split(",");
				//by saven. Apr. 13.
				x=Integer.parseInt(inner[0]);
				if(locHT.get(new Integer(x))==null)
				{
					locHT.put(new Integer(x), new Integer(groupid));
					groupid++;
					numGroups=groupid-1;
				}
				idHT.put(new Integer(nodeid),(Integer)(locHT.get(new Integer(x))));
				nodeid++;
			}

			in.close();
			for(int i=0;i<nodeid;i++)
			System.out.println("nodeid ="+i+", gid="+idHT.get(new Integer(i)));
			System.out.println("we have totally "+ numGroups+" groups.");
		}catch(Exception e)
		{
			System.out.println("we have an error:"+e);
		}


	}

	public int getGID(int x)
	{
		if(locHT.get(new Integer(x))==null)
		return -1;
		else
		return ((Integer)(locHT.get(new Integer(x)))).intValue();
	}
	public int getGIDbyID(int nodeid)
	{
			if(idHT.get(new Integer(nodeid))==null)
			return -1;
			else
			return ((Integer)(idHT.get(new Integer(nodeid)))).intValue();
	}
	public String doQuery(String query){
		String result="";

			try{
					//System.out.println("excuting query:"+query);
					Statement s = con.createStatement();
					s.execute(query);
					ResultSet rs = s.getResultSet(); // get any ResultSet that came from our query
					ResultSetMetaData mD=rs.getMetaData();
					int cN=mD.getColumnCount();
					int i=0;
					//System.out.println("cN="+cN);


					if (rs != null) // if rs == null, then there is no ResultSet to view
					while ( rs.next() ) // this will step through our data row-by-row
					{

						for(i=1;i<cN;i++)
						{
						String tmp=rs.getString(i);
						if(tmp!=null)
						result=result.concat(tmp+",");
						}
						String tmp=rs.getString(cN);
						if(tmp!=null)
						{
							//only for one value here, since we just have one value in aggregate!!
							//take care
							result=result.concat(tmp);
							break;
						}
						if(rs.next())
						result=result.concat(",");
						System.out.println(result);


					}
					s.close();
				}catch(Exception e)
				{
					System.out.println("error:"+e);
				}

		return result;
	}


	void importSensorData()
		{
			//* (1) build table with the name of directory
			File dic=new File(rdDir);
			File []files=dic.listFiles(new DatFilter());
			int step=(int)(SP/2);
			try
			{
				Statement s = con.createStatement();
				s.execute("create table sensordata (nodeid integer, epoch integer, temp integer, light integer, groupid integer)");
				for(int i=0;i<files.length;i++)
				{
					int epoch=0;
					int temp=0;
					int light=0;
					int nodeid=0;
					System.out.println(files[i].getName());
					nodeid=Integer.parseInt(files[i].getName().substring(6,7));
					int groupid=-1;
					/*if(nodeid== 5 || nodeid==4 || nodeid==8)
					groupid=1;
					else
					if(nodeid== 3 || nodeid==0 || nodeid==1)
					groupid=2;
					else
					if(nodeid== 7 || nodeid==2 || nodeid==6)
					groupid=3;*/
					//by saven. Apr 13
					groupid=getGIDbyID(nodeid);
					//System.out.println(nodeid);
					BufferedReader in = new BufferedReader(new FileReader(files[i].getPath()));
					String line=null;
					while(true)
					{
						line=in.readLine();
						for(int jjj=0;jjj<(step-1);jjj++)
						in.readLine();
						if(line==null) break;
						String []inner=line.split(",");
						//by saven. Apr. 13.
						temp=Integer.parseInt(inner[0]);
						light=Integer.parseInt(inner[1]);
						epoch++;
						String tuple="insert into sensordata values("+nodeid+", "+epoch+","+temp+","+light+","+groupid+")";
						System.out.println(tuple);
						s.execute(tuple);
					}

					in.close();
					con.commit();

				}
				s.close();


			}
				catch(Exception e)
				{
					System.out.println("Error in import log: " + e);
				}
	}
	void importLog()
	{
		importLog(aodvDir,"AODV");
		importLog(DDDir,"DD");
		importLog(tinyDBDir,"TinyDB");
	}
	void importLog(String directory,String className)
	{
		//* (1) build table with the name of directory
		File dic=new File(directory);
		File []subdirs=dic.listFiles(new LogQueryFilter());
	try
	{
		for(int i=0;i<subdirs.length;i++)
		{
			if(subdirs[i].isDirectory()==true)
			{

				String tablePrefix=className+"_"+subdirs[i].getName();
				System.out.println("quality file: "+tablePrefix);
				Statement s = con.createStatement();

				//s.execute("drop table "+tablePrefix);
				//s.execute("create table "+directory+" (nodeid integer, epoch integer, value1 integer, value2 integer)");
				s.execute("create table "+tablePrefix+" (epoch integer, values integer, groupid integer)");


				File []files=subdirs[i].listFiles(new LogFilter());
				int epoch=0;
				int value=0;
				int nodeid=0;
				boolean isSum=false;
				//if(tablePrefix.indexOf('7')!=-1 || tablePrefix.indexOf('8')!=-1)
				if(tablePrefix.indexOf('8')!=-1 || tablePrefix.indexOf('9')!=-1)
				isSum=false;
				else
				isSum=true;
				String line;
				String bigline;
				int delta18=0;
				int delta33=0;
				//we assume at most 10000 tuples for each sensor.
				int []epochArray=new int[10000];
				int []valueArray=new int[10000];
				int []groupidArray=new int[10000];
				int numValues=0;
				int groupid=-1;



				for(int j=0;j<files.length;j++)
				{
					System.out.println(files[j].getPath());
					BufferedReader in = new BufferedReader(new FileReader(files[j].getPath()));
					while(true)
					{
						bigline=in.readLine();
						if(bigline==null) break;
						//by saven. Apr.13.
						//String []newinin=bigline.split("  ");
						//System.out.println(newinin[0]+","+newinin[1]);
						//epoch=Integer.parseInt(newinin[0].substring(0,newinin[0].length()-1));
						line=bigline;
						String []inner=line.split(",");
						//filter the "0x7D 0x5D" to be "0x7D", "0x7D 0x5E" to be "0x7E"
						delta18=0;
						delta33=0;
						for(int k=0;k<(inner.length-1);k++)
						{
							if(Integer.parseInt(inner[k].substring(2),16)==0x7d && Integer.parseInt(inner[k+1].substring(2),16)==0x5d)
							{
								if(k<=9) delta18++;
								 else if(k<=36) delta33++;
								System.out.print("found: "+k);
							}
							if(Integer.parseInt(inner[k].substring(2),16)==0x7d && Integer.parseInt(inner[k+1].substring(2),16)==0x5e)
							{
								if(k<=18) delta18++;
								else if(k<=36) delta33++;
								System.out.print("found: "+k);
							}


						}
						if(delta33!=0)
						System.out.println("d33="+delta33);
						if(delta18!=0)
						System.out.println("d18="+delta18);



						epoch=Integer.parseInt(inner[16+delta18].substring(2),16);
						//nodeid=Integer.parseInt(inner[31+delta18].substring(2),16);
						int x=Integer.parseInt(inner[37+delta33].substring(2),16);
						System.out.print("we have: "+inner[37+delta33]+" ");
						System.out.println("N="+nodeid);
						if(isSum)
						{
							int low=Integer.parseInt(inner[33+delta33].substring(2),16);
							int high=Integer.parseInt(inner[34+delta33].substring(2),16);
							value=high*256+low;
						}
						else
						{
							/*int low1=Integer.parseInt(inner[33+delta33].substring(2),16);
							int high1=Integer.parseInt(inner[34+delta33].substring(2),16);
							int low2=Integer.parseInt(inner[35+delta33].substring(2),16);
							int high2=Integer.parseInt(inner[36+delta33].substring(2),16);*/
							int low1=Integer.parseInt(inner[33+delta33].substring(2),16);
							int high1=Integer.parseInt(inner[34+delta33].substring(2),16);
							int low2=Integer.parseInt(inner[35+delta33].substring(2),16);
							int high2=Integer.parseInt(inner[36+delta33].substring(2),16);
							int value1=high1*256+low1;
							int value2=high2*256+low2;
							//System.out.println(high1+","+high2+"-"+low1+","+low2+"~~"+value1+","+value2);
							if(value2>0)
							value=(int)(value1/value2);
							if(value<100)
							value=value1;

						}

						//dump the value to a large array.
						groupid=-1;
						/*if(nodeid== 5 || nodeid==4 || nodeid==8)
						groupid=1;
						else
						if(nodeid== 3 || nodeid==0 || nodeid==1)
						groupid=2;
						else
						if(nodeid== 7 || nodeid==2 || nodeid==6)
						groupid=3;*/
						//if(tablePrefix.indexOf('9')!=-1)
						groupid=getGID(x);
						epochArray[numValues]=epoch;
						valueArray[numValues]=value;
						groupidArray[numValues]=groupid;
						numValues++;



					}
					int numInsert=0;
					for(int mm=numValues-1;mm>=0;mm--)
					{
						//by saven. May06
						//if(valueArray[mm]<0||(groupidArray[mm]<0))
						if(valueArray[mm]<0)
						{
							String tuple="Error Tuple! value with "+valueArray[mm];
							System.out.println(tuple);
							continue;
						}

						String tuple="insert into "+tablePrefix+" values("+epochArray[mm]+", "+valueArray[mm]+","+groupidArray[mm]+")";
						System.out.println(tuple);
						s.execute(tuple);
						numInsert++;
						//by saven Apr. 13
						/*if((subdirs[i].getName().indexOf("9")==-1))
						{
							if(numInsert>=30) break;
						}
						else
						 	if(numInsert>=90) break;*/
					}
					con.commit();
					in.close();

					System.out.println("File: "+ files[j].getPath()+" okay.");
				}
				s.close();
			}
		}

	}
			catch(Exception e)
			{
				System.out.println("Error in import log: " +className+"," + e);
			}
	}

	/** function
	* Compute the data quality of the aggregation queries Q5-8
	* we set the valid values to be [0, 2*RV]
	**/
	void ComputeQuality()
	{
		ComputeQuality(aodvDir,"AODV");
		ComputeQuality(DDDir,"DD");
		ComputeQuality(tinyDBDir,"TinyDB");
	}
	public void ComputeQuality(String directory,String className)
	{
		File dic=new File(directory);
		File []subdirs=dic.listFiles(new LogQueryFilter());
		int m=0;
		int []tRV=new int[1000];
		try
		{
			PrintWriter out
				   = new PrintWriter(new BufferedWriter(new FileWriter(className+"quality.txt")));
			for(int i=0;i<subdirs.length;i++)
			{
				if(subdirs[i].isDirectory()==true)
				{
					for(m=5;m<=8;m++)
					{
						String temp;
						temp= m +"";
						if(subdirs[i].getName().indexOf(temp)!=-1)
						{
							m-=5;
							break;
						}
					}
					//for Q6,7,8
					if((subdirs[i].getName().indexOf("9")==-1))
					{
						//compute the real value into an array
						for(int tt=0;tt<1000;tt++)
						{
							String query=null;
							if(subdirs[i].getName().indexOf("6")!=-1)
								query="select max(temp) from sensordata where epoch ="+tt;
							   else
							    if(subdirs[i].getName().indexOf("7")!=-1)
							     	query="select sum(temp) from sensordata where epoch ="+tt;
							     	else
							     	if(subdirs[i].getName().indexOf("8")!=-1)
							     		query="select AVG(temp) from sensordata where temp > 400 and epoch="+tt;
							 String result=doQuery(query);
							 if(result.length()!=0)
							 	tRV[tt]=(int)Double.parseDouble(result);
							 else
							 	tRV[tt]=-1;
						}
						int tResult=0;
						int tReal=0;
						String tablePrefix=className+"_"+subdirs[i].getName();
						for(int tt=0;tt<1000;tt++)
						{
							if(tRV[tt]!=-1)
							{
								String query_1="select values from "+tablePrefix+" where values >"+tRV[tt] +" AND values <"+3*tRV[tt]+" and epoch ="+tt;
								String result_1=doQuery(query_1);

								String query_2="select values from "+tablePrefix+" where values <="+tRV[tt] +" AND values >0"+" and epoch ="+tt;
								String result_2=doQuery(query_2);
								if(result_1.length()!=0) //larger than RV
								{
									tResult=tResult+(int)Double.parseDouble(result_1)-tRV[tt];
									tReal=tReal+tRV[tt];
								}
								if(result_2.length()!=0) //larger than RV
								{
									tResult=tResult-(int)Double.parseDouble(result_2)+tRV[tt];
									tReal=tReal+tRV[tt];
								}

							}
						}
						double rate=(double)(tResult)/(double)(tReal);
						out.println(subdirs[i].getName()+","+rate);
						System.out.println(subdirs[i].getName()+", "+rate );
					}
					else
					{ //it is Q9, we have deal with 3 groups.
						out.println(subdirs[i].getName()+",");
						//by saven. Apr 13.
						for(int idid=1;idid<=numGroups;idid++)
						//for(int idid=1;idid<=3;idid++)
						{
							for(int tt=0;tt<1000;tt++)
							{
								String query=null;
								query="select AVG(temp) from sensordata where groupid = "+idid+" and epoch="+tt;
								String result=doQuery(query);
								if(result.length()!=0)
									tRV[tt]=(int)Double.parseDouble(result);
								else
									tRV[tt]=-1;
							}
							int tResult=0;
							int tReal=0;
							String tablePrefix=className+"_"+subdirs[i].getName();
							for(int tt=0;tt<1000;tt++)
							{
								if(tRV[tt]!=-1)
								{
									String query_1="select values from "+tablePrefix+" where values >"+tRV[tt] +" AND values <"+3*tRV[tt]+" and groupid = "+idid+" and epoch="+tt;
									String result_1=doQuery(query_1);

									String query_2="select values from "+tablePrefix+" where values <="+tRV[tt] +" AND values >0"+" and groupid = "+idid+" and epoch="+tt;
									String result_2=doQuery(query_2);
									//System.out.println(result_1+","+result_2);
									if(result_1.length()!=0) //larger than RV
									{
										tResult=tResult+(int)Double.parseDouble(result_1)-tRV[tt];
										tReal=tReal+tRV[tt];
									}
									if(result_2.length()!=0) //less than RV
									{
										tResult=tResult-(int)Double.parseDouble(result_2)+tRV[tt];
										tReal=tReal+tRV[tt];
									}

								}
							}
							System.out.println("tResult="+tResult+", tReal="+tReal);
							double rate=(double)(tResult)/(double)(tReal);
							System.out.println(subdirs[i].getName()+","+rate);
							out.println("G"+idid+", "+rate);
							/*String tablePrefix=className+"_"+subdirs[i].getName();
							String query_1="select sum(values-"+RV[i+idid-1]+"),count(epoch) from "+tablePrefix+" where values >"+RV[i+idid-1] +" AND values <"+2*RV[i+idid-1]+" AND groupid="+idid;
							String result_1=doQuery(query_1);
							String []inner_1=result_1.split(",");
							String query_2="select sum("+RV[i+idid-1]+"-values),count(epoch) from "+tablePrefix+" where values <"+RV[i+idid-1] +" AND values >0"+" AND groupid="+idid;
							String result_2=doQuery(query_2);
							String []inner_2=result_2.split(",");
							double result=0;
							if(inner_1[0].length()!=0 && inner_1[0]!=null)
							result=result+Double.parseDouble(inner_1[0]);
							if(inner_2[0].length()!=0 &&inner_2[0]!=null)
							result=result+Double.parseDouble(inner_2[0]);
							System.out.println(result);

							int count=0;
							if(inner_1[1].length()!=0 &&inner_1[1]!=null)
							count=count+Integer.parseInt(inner_1[1]);
							if(inner_2[1].length()!=0 &&inner_2[1]!=null)
							count=count+Integer.parseInt(inner_2[1]);
							double rate=(double)(result)/(double)(count*RV[i+idid-1]);
							System.out.println(subdirs[i].getName()+","+rate);
							out.println("G"+idid+", "+rate);*/
						}
					}
				}
			}
			out.close();
		}
		catch(Exception e)
		{
			System.out.println("Error in compute quality: " + e);
		}

	}
	public static void main(String []argv)
	{
		try {

					QualityAnalyzer log=new QualityAnalyzer();
					Class.forName("sun.jdbc.odbc.JdbcOdbcDriver");
					 // set this to a MS Access DB you have on your machine
					String filename = "QualityGeneral.mdb";
					String database = "jdbc:odbc:Driver={Microsoft Access Driver (*.mdb)};DBQ=";
					database+= filename.trim() + ";DriverID=22}"; // add on to the end
					 // now we can get the connection from the DriverManager
					log.con = DriverManager.getConnection( database ,"","");

					/*if (argv.length ==4)
					{
						log.aodvDir=argv[0];
						log.tinyDBDir=argv[1];
						log.DDDir=argv[2];
						log.sketchDir=argv[3];
						System.out.println("Getting quality information from: "+log.aodvDir+", "+log.tinyDBDir+", "+log.DDDir+", "+log.sketchDir);
						log.importLog();
						log.importSensorData();
						log.ComputeQuality();
					}
					else
					{
						System.out.println("error in arguments: java QualityAnalyzer <aodvdir> <tinydbdir> <dddir> <sketchdir> <RV1> <RV2> <RV3> <RV4>");
					}*/

					for(int i=0;i<argv.length;i++)
					{
						if(argv[i].equals("-Log"))
						{
							log.aodvDir=argv[i+1];
							log.tinyDBDir=argv[i+2];
							log.DDDir=argv[i+3];
							System.out.println("Getting quality information from: "+log.aodvDir+", "+log.tinyDBDir+", "+log.DDDir);
							log.importLog();
						}
						else
						if(argv[i].equals("-SP"))
						{
							log.SP=Integer.parseInt(argv[i+1]);
							System.out.println("SP="+log.SP);
						}
						else
						if(argv[i].equals("-RD"))
						{
							log.rdDir=argv[i+1];
							System.out.println("RD dir="+log.rdDir);
							log.importSensorData();

						}
						else
						if(argv[i].equals("-Loc"))
						{
							log.locFile=argv[i+1];
							log.buildLocHT();
						}


					}
					log.ComputeQuality();


					log.con.close(); // close the Connection to let the database know we're done with it*/


		        }
		            catch (Exception e) {
		            System.out.println("Error in main: " + e);
        }

	}

}